@@ -110,6 +110,12 @@ rest_proxy_call_dispose (GObject *object)
 {
   RestProxyCallPrivate *priv = GET_PRIVATE (object);
 
+  if (priv->cancellable)
+    {
+      g_signal_handler_disconnect (priv->cancellable, priv->cancel_sig);
+      g_clear_object (&priv->cancellable);
+    }
+
   if (priv->params)
   {
     rest_params_free (priv->params);
@@ -289,7 +295,7 @@ rest_proxy_call_add_header (RestProxyCall *call,
 /**
  * rest_proxy_call_add_headers:
  * @call: The #RestProxyCall
- * @Varargs: Header name and value pairs, followed by %NULL.
+ * @...: Header name and value pairs, followed by %NULL.
  *
  * Add the specified header name and value pairs to the call.  If a header
  * already exists, the new value will replace the old.
@@ -413,7 +419,7 @@ rest_proxy_call_add_param_full (RestProxyCall *call, RestParam *param)
 /**
  * rest_proxy_call_add_params:
  * @call: The #RestProxyCall
- * @Varargs: Parameter name and value pairs, followed by %NULL.
+ * @...: Parameter name and value pairs, followed by %NULL.
  *
  * Add the specified parameter name and value pairs to the call.  If a parameter
  * already exists, the new value will replace the old.
@@ -502,10 +508,11 @@ rest_proxy_call_remove_param (RestProxyCall *call,
  * rest_proxy_call_get_params:
  * @call: The #RestProxyCall
  *
- * Get the parameters as a #GHashTable of parameter names to values.  The caller
- * should call g_hash_table_unref() when they have finished using it.
+ * Get the parameters as a #GHashTable of parameter names to values.  The
+ * returned value is owned by the RestProxyCall instance and should not
+ * be freed by the caller.
  *
- * Returns: A #GHashTable.
+ * Returns: A #RestParams.
  */
 RestParams *
 rest_proxy_call_get_params (RestProxyCall *call)
@@ -602,7 +609,8 @@ finish_call (RestProxyCall *call, SoupMessage *message, GError **error)
       (SoupMessageHeadersForeachFunc)_populate_headers_hash_table,
       priv->response_headers);
 
-  priv->payload = g_strdup (message->response_body->data);
+  priv->payload = g_memdup (message->response_body->data,
+                            message->response_body->length + 1);
   priv->length = message->response_body->length;
 
   priv->status_code = message->status_code;
@@ -837,7 +845,7 @@ prepare_message (RestProxyCall *call, GError **error_out)
 }
 
 /**
- * rest_proxy_call_async: (skip):
+ * rest_proxy_call_async: (skip)
  * @call: The #RestProxyCall
  * @callback: a #RestProxyCallAsyncCallback to invoke on completion of the call
  * @weak_object: The #GObject to weakly reference and tie the lifecycle too
@@ -921,8 +929,6 @@ _call_message_call_completed_cb (SoupSession *session,
   call = REST_PROXY_CALL (
       g_async_result_get_source_object (G_ASYNC_RESULT (result)));
 
-  // FIXME: disconnect cancellable ?
-
   finish_call (call, message, &error);
 
   if (error != NULL)
@@ -973,8 +979,11 @@ rest_proxy_call_invoke_async (RestProxyCall      *call,
                                       user_data, rest_proxy_call_invoke_async);
 
   if (cancellable != NULL)
-    g_signal_connect (cancellable, "cancelled",
-        G_CALLBACK (_call_message_call_cancelled_cb), call);
+    {
+      priv->cancel_sig = g_signal_connect (cancellable, "cancelled",
+          G_CALLBACK (_call_message_call_cancelled_cb), call);
+      priv->cancellable = g_object_ref (cancellable);
+    }
 
   _rest_proxy_queue_message (priv->proxy,
                              message,
@@ -1026,12 +1035,12 @@ _continuous_call_message_got_chunk_cb (SoupMessage                    *msg,
 
 
 /**
- * rest_proxy_call_continuous:
+ * rest_proxy_call_continuous: (skip)
  * @call: The #RestProxyCall
  * @callback: a #RestProxyCallContinuousCallback to invoke when data is available
  * @weak_object: The #GObject to weakly reference and tie the lifecycle to
- * @userdata: data to pass to @callback
- * @error: a #GError, or %NULL
+ * @userdata: (closure): data to pass to @callback
+ * @error: (out) (allow-none): a #GError, or %NULL
  *
  * Asynchronously invoke @call but expect a continuous stream of content. This
  * means that the body data will not be accumulated and thus you cannot use
@@ -1164,8 +1173,8 @@ _upload_call_message_wrote_data_cb (SoupMessage                *msg,
 /**
  * rest_proxy_call_upload:
  * @call: The #RestProxyCall
- * @callback: a #RestProxyCallUploadCallback to invoke when a chunk of data was
- *   uploaded
+ * @callback: (scope async): a #RestProxyCallUploadCallback to invoke when a chunk
+ *   of data was uploaded
  * @weak_object: The #GObject to weakly reference and tie the lifecycle to
  * @userdata: data to pass to @callback
  * @error: a #GError, or %NULL
@@ -1239,7 +1248,7 @@ rest_proxy_call_upload (RestProxyCall                *call,
 }
 
 /**
- * rest_proxy_call_cancel: (skip):
+ * rest_proxy_call_cancel: (skip)
  * @call: The #RestProxyCall
  *
  * Cancel this call.  It may be too late to not actually send the message, but
@@ -1259,6 +1268,12 @@ rest_proxy_call_cancel (RestProxyCall *call)
   priv = GET_PRIVATE (call);
   closure = priv->cur_call_closure;
 
+  if (priv->cancellable)
+    {
+      g_signal_handler_disconnect (priv->cancellable, priv->cancel_sig);
+      g_clear_object (&priv->cancellable);
+    }
+
   if (closure)
   {
     /* This will cause the _call_message_completed_cb to be fired which will
