@@ -28,6 +28,8 @@
 #include <libsoup/soup-gnome.h>
 #endif
 
+#include "rest-marshal.h"
+#include "rest-proxy-auth-private.h"
 #include "rest-proxy.h"
 #include "rest-private.h"
 
@@ -42,10 +44,13 @@ struct _RestProxyPrivate {
   gchar *url_format;
   gchar *url;
   gchar *user_agent;
+  gchar *username;
+  gchar *password;
   gboolean binding_required;
   SoupSession *session;
   SoupSession *session_sync;
   gboolean disable_cookies;
+  char *ssl_ca_file;
 };
 
 enum
@@ -54,9 +59,21 @@ enum
   PROP_URL_FORMAT,
   PROP_BINDING_REQUIRED,
   PROP_USER_AGENT,
-  PROP_DISABLE_COOKIES
+  PROP_DISABLE_COOKIES,
+  PROP_USERNAME,
+  PROP_PASSWORD,
+  PROP_SSL_STRICT,
+  PROP_SSL_CA_FILE
 };
 
+enum {
+  AUTHENTICATE,
+  LAST_SIGNAL
+};
+
+static guint signals[LAST_SIGNAL] = { 0 };
+
+
 static gboolean _rest_proxy_simple_run_valist (RestProxy *proxy, 
                                                char     **payload, 
                                                goffset   *len,
@@ -95,6 +112,24 @@ rest_proxy_get_property (GObject   *object,
     case PROP_DISABLE_COOKIES:
       g_value_set_boolean (value, priv->disable_cookies);
       break;
+    case PROP_USERNAME:
+      g_value_set_string (value, priv->username);
+      break;
+    case PROP_PASSWORD:
+      g_value_set_string (value, priv->password);
+      break;
+    case PROP_SSL_STRICT: {
+      gboolean ssl_strict;
+      g_object_get (G_OBJECT(priv->session),
+                    "ssl-strict", &ssl_strict,
+                    NULL);
+      g_value_set_boolean (value, ssl_strict);
+      break;
+    }
+    case PROP_SSL_CA_FILE:
+      g_value_set_string (value, priv->ssl_ca_file);
+      break;
+
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
   }
@@ -131,6 +166,26 @@ rest_proxy_set_property (GObject      *object,
     case PROP_DISABLE_COOKIES:
       priv->disable_cookies = g_value_get_boolean (value);
       break;
+    case PROP_USERNAME:
+      g_free (priv->username);
+      priv->username = g_value_dup_string (value);
+      break;
+    case PROP_PASSWORD:
+      g_free (priv->password);
+      priv->password = g_value_dup_string (value);
+      break;
+    case PROP_SSL_STRICT:
+      g_object_set (G_OBJECT(priv->session),
+                    "ssl-strict", g_value_get_boolean (value),
+                    NULL);
+      g_object_set (G_OBJECT(priv->session_sync),
+                    "ssl-strict", g_value_get_boolean (value),
+                    NULL);
+      break;
+    case PROP_SSL_CA_FILE:
+      g_free(priv->ssl_ca_file);
+      priv->ssl_ca_file = g_value_dup_string (value);
+      break;
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
   }
@@ -156,6 +211,36 @@ rest_proxy_dispose (GObject *object)
   G_OBJECT_CLASS (rest_proxy_parent_class)->dispose (object);
 }
 
+static gboolean
+default_authenticate_cb (RestProxy *self,
+                         G_GNUC_UNUSED RestProxyAuth *auth,
+                         gboolean retrying)
+{
+  /* We only want to try the credentials once, otherwise we get in an
+   * infinite loop with failed credentials, retrying the same invalid
+   * ones again and again
+   */
+  return !retrying;
+}
+
+static void
+authenticate (RestProxy   *self,
+              SoupMessage *msg,
+              SoupAuth    *soup_auth,
+              gboolean     retrying,
+              SoupSession *session)
+{
+  RestProxyPrivate *priv = GET_PRIVATE (self);
+  RestProxyAuth *rest_auth;
+  gboolean try_auth;
+
+  rest_auth = rest_proxy_auth_new (self, session, msg, soup_auth);
+  g_signal_emit(self, signals[AUTHENTICATE], 0, rest_auth, retrying, &try_auth);
+  if (try_auth && !rest_proxy_auth_is_paused (rest_auth))
+    soup_auth_authenticate (soup_auth, priv->username, priv->password);
+  g_object_unref (G_OBJECT (rest_auth));
+}
+
 static void
 rest_proxy_constructed (GObject *object)
 {
@@ -168,6 +253,22 @@ rest_proxy_constructed (GObject *object)
     soup_session_add_feature (priv->session_sync, cookie_jar);
     g_object_unref (cookie_jar);
   }
+
+  if (REST_DEBUG_ENABLED(PROXY)) {
+    SoupSessionFeature *logger = (SoupSessionFeature*)soup_logger_new (SOUP_LOGGER_LOG_BODY, 0);
+    soup_session_add_feature (priv->session, logger);
+    g_object_unref (logger);
+
+    logger = (SoupSessionFeature*)soup_logger_new (SOUP_LOGGER_LOG_BODY, 0);
+    soup_session_add_feature (priv->session_sync, logger);
+    g_object_unref (logger);
+  }
+
+  /* session lifetime is same as self, no need to keep signalid */
+  g_signal_connect_swapped (priv->session, "authenticate",
+                            G_CALLBACK(authenticate), object);
+  g_signal_connect_swapped (priv->session_sync, "authenticate",
+                            G_CALLBACK(authenticate), object);
 }
 
 static void
@@ -178,6 +279,9 @@ rest_proxy_finalize (GObject *object)
   g_free (priv->url);
   g_free (priv->url_format);
   g_free (priv->user_agent);
+  g_free (priv->username);
+  g_free (priv->password);
+  g_free (priv->ssl_ca_file);
 
   G_OBJECT_CLASS (rest_proxy_parent_class)->finalize (object);
 }
@@ -238,6 +342,78 @@ rest_proxy_class_init (RestProxyClass *klass)
   g_object_class_install_property (object_class,
                                    PROP_DISABLE_COOKIES,
                                    pspec);
+
+  pspec = g_param_spec_string ("username",
+                               "username",
+                               "The username for authentication",
+                               NULL,
+                               G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class,
+                                   PROP_USERNAME,
+                                   pspec);
+
+  pspec = g_param_spec_string ("password",
+                               "password",
+                               "The password for authentication",
+                               NULL,
+                               G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class,
+                                   PROP_PASSWORD,
+                                   pspec);
+
+  pspec = g_param_spec_boolean ("ssl-strict",
+                                "Strictly validate SSL certificates",
+                                "Whether certificate errors should be considered a connection error",
+                                TRUE,
+                                G_PARAM_READWRITE);
+  g_object_class_install_property (object_class,
+                                   PROP_SSL_STRICT,
+                                   pspec);
+
+  pspec = g_param_spec_string ("ssl-ca-file",
+                               "SSL CA file",
+                               "File containing SSL CA certificates.",
+                               NULL,
+                               G_PARAM_READWRITE);
+  g_object_class_install_property (object_class,
+                                   PROP_SSL_CA_FILE,
+                                   pspec);
+
+  /**
+   * RestProxy::authenticate:
+   * @proxy: the proxy
+   * @auth: authentication state
+   * @retrying: %TRUE if this is the second (or later) attempt
+   *
+   * Emitted when the proxy requires authentication. If
+   * credentials are available, set the 'username' and 'password'
+   * properties on @proxy and return %TRUE from the callback.
+   * This will cause the signal emission to stop, and librest will
+   * try to connect with these credentials
+   * If these credentials fail, the signal will be
+   * emitted again, with @retrying set to %TRUE, which will
+   * continue until %FALSE is returned from the callback.
+   *
+   * If you call rest_proxy_auth_pause() on @auth before
+   * returning, then you can the authentication credentials on
+   * the #RestProxy object asynchronously. You have to make sure
+   * that @auth does not get destroyed with g_object_ref().
+   * You can then unpause the authentication with
+   * rest_proxy_auth_unpause() when everything is ready for it
+   * to continue.
+   **/
+  signals[AUTHENTICATE] =
+      g_signal_new ("authenticate",
+                    G_OBJECT_CLASS_TYPE (object_class),
+                    G_SIGNAL_RUN_LAST,
+                    G_STRUCT_OFFSET (RestProxyClass, authenticate),
+                    g_signal_accumulator_true_handled, NULL,
+                    g_cclosure_user_marshal_BOOLEAN__OBJECT_BOOLEAN,
+                    G_TYPE_BOOLEAN, 2,
+                    REST_TYPE_PROXY_AUTH,
+                    G_TYPE_BOOLEAN);
+
+  proxy_class->authenticate = default_authenticate_cb;
 }
 
 static void
@@ -248,6 +424,7 @@ rest_proxy_init (RestProxy *self)
   priv->session = soup_session_async_new ();
   priv->session_sync = soup_session_sync_new ();
 
+#ifdef REST_SYSTEM_CA_FILE
   /* with ssl-strict (defaults TRUE) setting ssl-ca-file forces all
    * certificates to be trusted */
   g_object_set (priv->session,
@@ -256,6 +433,13 @@ rest_proxy_init (RestProxy *self)
   g_object_set (priv->session_sync,
                 "ssl-ca-file", REST_SYSTEM_CA_FILE,
                 NULL);
+#endif
+  g_object_bind_property (self, "ssl-ca-file",
+                          priv->session, "ssl-ca-file",
+                          G_BINDING_BIDIRECTIONAL);
+  g_object_bind_property (self, "ssl-ca-file",
+                          priv->session_sync, "ssl-ca-file",
+                          G_BINDING_BIDIRECTIONAL);
 
 #if WITH_GNOME
   soup_session_add_feature_by_type (priv->session,
@@ -263,16 +447,6 @@ rest_proxy_init (RestProxy *self)
   soup_session_add_feature_by_type (priv->session_sync,
                                     SOUP_TYPE_PROXY_RESOLVER_GNOME);
 #endif
-
-  if (REST_DEBUG_ENABLED(PROXY)) {
-    SoupSessionFeature *logger = (SoupSessionFeature*)soup_logger_new (SOUP_LOGGER_LOG_BODY, 0);
-    soup_session_add_feature (priv->session, logger);
-    g_object_unref (logger);
-
-    logger = (SoupSessionFeature*)soup_logger_new (SOUP_LOGGER_LOG_BODY, 0);
-    soup_session_add_feature (priv->session_sync, logger);
-    g_object_unref (logger);
-  }
 }
 
 /**
@@ -290,15 +464,45 @@ rest_proxy_init (RestProxy *self)
  * Returns: A new #RestProxy.
  */
 RestProxy *
-rest_proxy_new (const gchar *url_format, 
+rest_proxy_new (const gchar *url_format,
                 gboolean     binding_required)
 {
-  return g_object_new (REST_TYPE_PROXY, 
+  return g_object_new (REST_TYPE_PROXY,
                        "url-format", url_format,
                        "binding-required", binding_required,
                        NULL);
 }
 
+/**
+ * rest_proxy_new_with_authentication:
+ * @url_format: the endpoint URL
+ * @binding_required: whether the URL needs to be bound before calling
+ * @username: the username provided by the user or client
+ * @password: the password provided by the user or client
+ *
+ * Create a new #RestProxy for the specified endpoint @url_format, using the
+ * "GET" method.
+ *
+ * Set @binding_required to %TRUE if the URL contains string formatting
+ * operations (for example "http://foo.com/%<!-- -->s".  These must be expanded
+ * using rest_proxy_bind() before invoking the proxy.
+ *
+ * Returns: A new #RestProxy.
+ */
+RestProxy *
+rest_proxy_new_with_authentication (const gchar *url_format,
+                                    gboolean     binding_required,
+                                    const gchar *username,
+                                    const gchar *password)
+{
+  return g_object_new (REST_TYPE_PROXY,
+                       "url-format", url_format,
+                       "binding-required", binding_required,
+                       "username", username,
+                       "password", password,
+                       NULL);
+}
+
 static gboolean
 _rest_proxy_bind_valist (RestProxy *proxy,
                          va_list    params)
@@ -374,6 +578,15 @@ _rest_proxy_new_call (RestProxy *proxy)
   return call;
 }
 
+/**
+ * rest_proxy_new_call:
+ * @proxy: the #RestProxy
+ *
+ * Create a new #RestProxyCall for making a call to the web service.  This call
+ * is one-shot and should not be re-used for making multiple calls.
+ *
+ * Returns: (transfer full): a new #RestProxyCall.
+ */
 RestProxyCall *
 rest_proxy_new_call (RestProxy *proxy)
 {
